// -*- C++ -*-
/**
        @file       hotOcean.cpp
        @since      2013-09-30

        @author     Nico Rehberg

        @brief      Mtoa shader to color a surface using Houdini Ocean Toolkit (implementation).

        Implementation: Nico Rehberg <mail@nico-rehberg.de>

        The Houdini Ocean Toolkit is copyrighted by Drew Whitehouse
        see http://odforce.net/wiki/index.php/HoudiniOceanToolkit

        This program is free software; you can redistribute it and/or
        modify it under the terms of the GNU General Public License
        as published by the Free Software Foundation; either version 2
        of the License, or (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.
*/


#include <ai.h>
#include <string.h>
#include "math.h"
#include <iostream>

#include "mtoaOcean.h"
#ifndef M_PI
#define M_PI           3.14159265358979323846
#endif

// Shader Data Structure
//
typedef struct 
{
    mtoaOcean *ocean;
    int deformSpace;   
    int outputMode; 
    float fov;
    int xres;
    float oneOverGlobalScale;
}
ShaderData;


AI_SHADER_NODE_EXPORT_METHODS(hotOceanSimpleMethods);

enum hotOceanSimpleParams {p_globalScale, p_resolution, p_size, p_windSpeed, p_waveHeight, p_shortestWave, p_choppiness, p_windDirection, \
p_dampReflections, p_windAlign, p_oceanDepth, p_time, p_seed, p_deformSpace, p_outputMode, p_inDisplacement};

enum deformSpaceModes {WORLD, OBJECT, TANGENT};
const char* deformSpaceModesNames1[] = {"world", "object", "tangent", NULL};

enum outputModes {VDISP, JMINUS, JPLUS, EMINUS, EPLUS};
const char* outputModesNames[] = {"vectorDisplacement", "jMinus", "jPlus", "EMinus", "EPlus", NULL};

node_parameters
{

    AiMetaDataSetStr(nentry, NULL, "maya.name","aiHotOcean");
    AiMetaDataSetInt(nentry, NULL, "maya.id",0x0011CF82);
                                            

    AiMetaDataSetStr(nentry, NULL, "maya.classification", "texture/utility");
    AiMetaDataSetStr(nentry, NULL, "maya.output_name", "outColor");
    AiMetaDataSetStr(nentry, NULL, "maya.output_shortname", "out");

    AiParameterFlt ("globalScale"   , 1.0);
    AiParameterInt ("resolution"    , 8);	// grid aufloesung
    AiParameterFlt ("size"          , 200);     // The grid mentiond above is computed for and applied to the input geometry in tiles of this size.
    AiParameterFlt ("windSpeed"     , 30);      // Wind Speed - Affects the shape of the waves, "Windspeed (m/s)"
    AiParameterFlt ("waveHeight"    , 1);
    AiParameterFlt ("shortestWave"  , 0.02f);    // Shortest Wavelength(m)
    AiParameterFlt ("choppiness"    , 0.8f);
    AiParameterFlt ("windDirection" , 0);       // Wind direction in degrees
    AiParameterFlt ("dampReflections", 0.5);    // Damp reflections - In a fully developed ocean you will have waves travelling in both the forward and backwards directions. This parameter damps out the negative direcion waves.
    AiParameterFlt ("windAlign"     , 2.0);     // Wind Alignment - Controls how closely the waves travel in the direction of the wind.
    AiParameterFlt ("oceanDepth"    , 200);     // Ocean Depth - Affects the spectrum of waves generated. Visually in doesn't seem to have that great an influence.
    AiParameterFlt ("time"          , 0);
    AiParameterInt ("seed"          , 0);       // Seed - Seeds the random number generator.
    AiParameterEnum ("deformSpace"   , 0, deformSpaceModesNames1);  // world - object - tangent
    AiParameterEnum ("outputMode"    , 0, outputModesNames);  // vDisp, jminus, jplus ..
    AiParameterVec ("inDisplacement", 0, 0, 0); // an input for layering multiple displacements
}


node_initialize
{
	//ShaderData *data = (ShaderData*) AiMalloc(sizeof(ShaderData));
	ShaderData* data = new ShaderData;
    AiNodeSetLocalData(node,data); 
    data->ocean = new mtoaOcean(); 
    AiMsgInfo("Ocean init");
}

node_update
{
    ShaderData *data = (ShaderData*)AiNodeGetLocalData(node);    
    data->deformSpace = AiNodeGetInt(node, "deformSpace");
    data->outputMode = AiNodeGetInt(node, "outputMode");
    data->oneOverGlobalScale = 1.0f/ AiNodeGetFlt(node, "globalScale");
    bool do_chop = false;
    if (AiNodeGetFlt(node, "choppiness") > 0.0001)
        do_chop = true;    

    bool do_jacobian = (data->outputMode > 0);
    bool do_normal = false;
    
    int res = AiNodeGetInt(node, "resolution");
    res = (int)pow (2.0, res);

    delete (data->ocean->_ocean_context); 
    delete (data->ocean->_ocean);

    data->ocean->_ocean = new drw::Ocean(res, res, AiNodeGetFlt(node, "size")/float(res), AiNodeGetFlt(node, "size")/float(res),
                                AiNodeGetFlt(node, "windSpeed"), AiNodeGetFlt(node, "shortestWave"), 0.00001f, AiNodeGetFlt(node, "windDirection")/180.0f * M_PI,
                                1.0f - AiNodeGetFlt(node, "dampReflections"), AiNodeGetFlt(node, "windAlign"), AiNodeGetFlt(node, "oceanDepth"), AiNodeGetInt(node, "seed"));
    data->ocean->_normalize_factor = data->ocean->_ocean->get_height_normalize_factor();
    
    data->ocean->_ocean_context = data->ocean->_ocean->new_context(true, do_chop, do_normal, do_jacobian);
    
    // sum up the waves at this timestep
    data->ocean->_ocean->update(AiNodeGetFlt(node, "time"), *data->ocean->_ocean_context, true, do_chop, do_normal, do_jacobian, data->ocean->_normalize_factor * AiNodeGetFlt(node, "waveHeight"), AiNodeGetFlt(node, "choppiness"));
    
    AiMsgInfo(" ");
    AiMsgInfo("HotOcean %s",  AiNodeGetName(node));
    AiMsgInfo("-------------------------");
    AiMsgInfo("DeformSpace = %s", deformSpaceModesNames1[data->deformSpace]);
    AiMsgInfo("OutputMode = %s", outputModesNames[data->outputMode]);
    AiMsgInfo("Normalisation factor = %f", data->ocean->_normalize_factor);
    AiMsgInfo("Time = %f", AiNodeGetFlt(node, "time"));
}

shader_evaluate
{
    ShaderData *data = (ShaderData*)AiNodeGetLocalData(node);
    //AtParamValue *params = AiNodeGetParams(node);    
    AtVector disp, inDisp;
    inDisp = AiShaderEvalParamVec(p_inDisplacement);
    drw::EvalData evaldata;

    switch(data->deformSpace)
    {
    case WORLD: 
        {
        AtVector inDispWs = AiM4VectorByMatrixMult(sg->M, inDispWs);
        data->ocean->_ocean_context->eval2_xz(data->oneOverGlobalScale*(sg->P.x + inDispWs.x), data->oneOverGlobalScale*(sg->P.z + inDispWs.z), evaldata);    
        disp.x = evaldata.disp[0] * AiShaderEvalParamFlt(p_globalScale);
        disp.y = evaldata.disp[1] * AiShaderEvalParamFlt(p_globalScale);
        disp.z = evaldata.disp[2] * AiShaderEvalParamFlt(p_globalScale);
        //simple approximation to get the point before deformation to calculate the utility channels
        if (data->outputMode > 0) 
            data->ocean->_ocean_context->eval2_xz(data->oneOverGlobalScale*(sg->P.x - disp.x), data->oneOverGlobalScale*(sg->P.z - disp.z), evaldata);
        disp = AiM4VectorByMatrixMult(sg->Minv, disp);
        break;
        }
    case OBJECT: 
        {		
        data->ocean->_ocean_context->eval2_xz(data->oneOverGlobalScale*(sg->Po.x + inDisp.x), data->oneOverGlobalScale*(sg->Po.z + inDisp.z), evaldata);		
        disp.x = evaldata.disp[0] * AiShaderEvalParamFlt(p_globalScale);
        disp.y = evaldata.disp[1] * AiShaderEvalParamFlt(p_globalScale);
        disp.z = evaldata.disp[2] * AiShaderEvalParamFlt(p_globalScale);		
        //simple approximation to get the point before deformation to calculate the utility channels
        if (data->outputMode > 0) 
            data->ocean->_ocean_context->eval2_xz(data->oneOverGlobalScale*(sg->Po.x - disp.x), data->oneOverGlobalScale*(sg->Po.z - disp.z), evaldata);
        break;
        }
    case TANGENT:
        {
        data->ocean->_ocean_context->eval2_xz(data->oneOverGlobalScale*sg->u, data->oneOverGlobalScale*sg->v, evaldata);
        AtVector tangent = -1.0 * AiV3Normalize(sg->dPdx);
        AtVector biNormal = -1.0 * AiV3Normalize(sg->dPdy);
        disp = evaldata.disp[0] * AiShaderEvalParamFlt(p_globalScale) * tangent;
        disp += evaldata.disp[1] * AiShaderEvalParamFlt(p_globalScale) * sg->N;
        disp += evaldata.disp[2] * AiShaderEvalParamFlt(p_globalScale) * biNormal;
        // AiM4VectorByMatrixMult(&disp, sg->M, &disp);   
        break;
        }
    }

    switch(data->outputMode)
    {
    case VDISP:
        disp += inDisp;
		//sg->out.RGB() = AI_RGB_50GREY;
		//AiMsgInfo("%f %f %f", disp.x, disp.y, disp.z);
        sg->out.RGB().r = disp.x;
        sg->out.RGB().g = disp.y;
        sg->out.RGB().b = disp.z;
        break;
    case JMINUS:
        sg->out.RGB().r = evaldata.Jminus;
        sg->out.RGB().g = evaldata.Jminus;
        sg->out.RGB().b = evaldata.Jminus;
        break;
    case JPLUS:
        sg->out.RGB().r = evaldata.Jplus;
        sg->out.RGB().g = evaldata.Jplus;
        sg->out.RGB().b = evaldata.Jplus;
        break;
    case EMINUS:
        sg->out.RGB().r = evaldata.Eminus[0];
        sg->out.RGB().g = evaldata.Eminus[1];
        sg->out.RGB().b = evaldata.Eminus[2];
        break;
    case EPLUS:
        sg->out.RGB().r = evaldata.Eplus[0];
        sg->out.RGB().g = evaldata.Eplus[1];
        sg->out.RGB().b = evaldata.Eplus[2];
        break;
    }
    
}


node_finish
{
    ShaderData *data = (ShaderData*)AiNodeGetLocalData(node);
    delete (data->ocean->_ocean_context); 
    delete (data->ocean->_ocean);
    AiFree(data);
}

node_loader
{
   if (i > 0) 
        return 0;
   
   node->methods      = hotOceanSimpleMethods;
   node->output_type  = AI_TYPE_RGB;
   node->name         = "hotOcean";
   node->node_type    = AI_NODE_SHADER;
   strcpy(node->version, AI_VERSION);
   return 1;
}

