// -*- C++ -*-
/**
        @file       VRayHotOcean.cpp
        @since      2015-09-24

        @author     Nico Rehberg

        @brief      Vray shader to color a surface using Houdini Ocean Toolkit (implementation).

        Implementation: Nico Rehberg <mail@nico-rehberg.de>

        The Houdini Ocean Toolkit is copyrighted by Drew Whitehouse
        see http://odforce.net/wiki/index.php/HoudiniOceanToolkit

        This program is free software; you can redistribute it and/or
        modify it under the terms of the GNU General Public License
        as published by the Free Software Foundation; either version 2
        of the License, or (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.
*/

#include "vrayplugins.h"
#include "vrayinterface.h"
#include "vrayrenderer.h"
#include "brdfs.h"
#include "vraytexutils.h"

#include "globalnewdelete.cpp"

#include "mtoaOcean.h"

using namespace VR;  

#ifndef M_PI
#define M_PI           3.14159265358979323846
#endif

struct TexmapCache {
	AColor color;
	VR::Color jMinus, jPlus, Eminus, Eplus;
};

// Parameters
struct TexHotOcean_Params: VRayParameterListDesc {
	TexHotOcean_Params(void) {
		// Add the parameters
		addParamFloat		("globalScale",		1.0f,	-1,		"");
		addParamInt			("resolution",		8,		-1,		"Grid Resolution in powers of 2");
		addParamFloat		("size",			200.0f,	-1,		"The grid mentiond above is computed for and applied to the input geometry in tiles of this size.");
		addParamFloat		("windSpeed",		30.0f,	-1,		"Wind Speed - Affects the shape of the waves, Windspeed (m/s)");
		addParamFloat		("waveHeight",		1.0f,	-1,		"");
		addParamFloat		("shortestWave",	0.02f,	-1,		"Shortest Wavelength(m)");
		addParamFloat		("choppiness",		0.8f,	-1,		"");
		addParamFloat		("windDirection",	0.0f,	-1,		"Wind direction in degrees");
		addParamFloat		("dampReflections",	0.5f,	-1,		"Damp reflections - In a fully developed ocean you will have waves travelling in both the forward and backwards directions. This parameter damps out the negative direcion waves.");
		addParamFloat		("windAlign",		2.0f,	-1,		"Wind Alignment - Controls how closely the waves travel in the direction of the wind.");
		addParamFloat		("oceanDepth",		200.0f,	-1,		"Ocean Depth - Affects the spectrum of waves generated. Visually in doesn't seem to have that great an influence.");
		addParamFloat		("time",			1.0f,	-1,		"");
		addParamInt			("seed",			0,		-1,		"Seeds the random number generator.");
		addParamInt			("deformSpace",		0,		-1,		"Space in which to calc the deformation (0 - world; 1 - object; 2 - tangent;)", "enum=(0:world; 1:object; 2:tangent;)");
		addParamInt			("outputMode",		0,		-1,		"what output should the shader generate (0 - Vector Displacement; 1 - jMinus; 2 - jPlus; 3 - eMinus; 4 - ePlus)", "enum=(0:vectorDisplacement; 1:jMinus; 2:jPlus; 3:eMinus; 4:ePlus)");
		addParamTexture		("inDisplacement",			-1 ,	"an input for layering multiple displacements");

		addOutputParamTexture("eMinus", "eMinus.");
		addOutputParamTexture("ePlus", "ePlus.");
		addOutputParamTextureFloat("jMinus", "Jminus.");
		addOutputParamTextureFloat("jPlus", "JPLus.");

	}
};

struct TexHotOcean : VRayTexture, RTTexInterface {
	TexHotOcean(VRayPluginDesc *pluginDesc):VRayTexture(pluginDesc) {
		// Set parameter caches
		paramList->setParamCache("globalScale", &globalScale);
		paramList->setParamCache("resolution", &resolution);
		paramList->setParamCache("size", &size);
		paramList->setParamCache("windSpeed", &windSpeed);
		paramList->setParamCache("waveHeight", &waveHeight);
		paramList->setParamCache("shortestWave", &shortestWave);
		paramList->setParamCache("choppiness", &choppiness);
		paramList->setParamCache("windDirection", &windDirection);
		paramList->setParamCache("dampReflections", &dampReflections);
		paramList->setParamCache("windAlign", &windAlign);
		paramList->setParamCache("oceanDepth", &oceanDepth);
		paramList->setParamCache("time", &time);
		paramList->setParamCache("seed", &seed);
		paramList->setParamCache("deformSpace", &deformSpace);
		paramList->setParamCache("outputMode", &outputMode);
		paramList->setParamCache("inDisplacement", &inDisplacement);

		//paramList->setOutputParamCache("eMinus", static_cast<TextureInterface*>(this));
		//paramList->setOutputParamCache("ePlus", static_cast<TextureInterface*>(this));
		//paramList->setOutputParamCache("jMinus", static_cast<TextureFloatInterface*>(&jMinus));
		//paramList->setOutputParamCache("jPlus", static_cast<TextureFloatInterface*>(this));
	}

	void renderBegin(VR::VRayRenderer *vray);
	void renderEnd(VR::VRayRenderer *vray);
	void frameBegin(VR::VRayRenderer *vray);
	void frameEnd(VR::VRayRenderer *vray);

	//these need to be reimplementet in a texture node
	AColor getTexColor(const VRayContext &rc);
	void getTexColorBounds(AColor &cmin, AColor &cmax);
	Vector getColorBumpGradient(const VRayContext &rc);

	PluginBase* getPlugin(void) { return (PluginBase*) this; }
private:
	void writeElements(const VR::VRayContext &rc, const TexmapCache &res);
	// Cached parameters
	TextureInterface *inDisplacement;	
	float globalScale, windSpeed, waveHeight, shortestWave, choppiness, windDirection, dampReflections, windAlign, oceanDepth, size, time;
	int resolution, seed, deformSpace, outputMode;
	float jMinus, jPlus;

	//cached ocean stuff
	mtoaOcean *ocean;
    float oneOverGlobalScale;

	// The color cache
	ShadeCache<TexmapCache, true, true> shadeCache;

	// from RTTexInterface
	PluginInterface *newInterface( InterfaceID id )
	{
		if( id == EXT_RTTEX )
		{
			return static_cast<PluginInterface*>( (RTTexInterface*)this );
		}

		return VRayTexture::newInterface(id);
	}

	virtual int getDataSize(void) { return 0; }
	virtual void getData(float *fdata) { }
	virtual int getSource(const char *part,const char *outParam,const char *lang,VUtils::ShaderSource &, OCLData &mtlData);

	/// Prints the Shader source of a texture parameter for the TexHotOcean plug-in if there is no texture it prints black color
	/// @param component the texture parameter
	/// @param part the part of the source code
	/// @param outParam deprecated parameter
	/// @param src the Shader source so far
	/// @param mtlData RT GPU material data
	int printComponentSource(PluginBase* component, const char *part,const char *outParam,const char *lang,VUtils::ShaderSource &src, OCLData &mtlData);
};

#define TexHotOcean_PluginID PluginID(LARGE_CONST(2024920329))
PLUGIN_DESC(TexHotOcean_PluginID, EXT_TEXTURE, "TexHotOcean", TexHotOcean, TexHotOcean_Params);
PLUGIN_LIBRARY("TexHotOcean", "HotOcean texture.");

void TexHotOcean::renderBegin(VR::VRayRenderer *vray) {
	VRayTexture::renderBegin(vray);
	shadeCache.renderBegin(vray);
}

void TexHotOcean::renderEnd(VR::VRayRenderer *vray) {
	VRayTexture::renderEnd(vray);
	shadeCache.renderEnd(vray);
}

void TexHotOcean::frameBegin(VR::VRayRenderer *vray) {
	// Call the base class to update the parameter caches
	VRayTexture::frameBegin(vray);
	// Call the shade cache
	shadeCache.frameBegin(vray);

	//set up the ocean
	ocean = new mtoaOcean();
    oneOverGlobalScale = 1.0/globalScale;    
    bool do_chop = false;
    if ( choppiness > 0.0001)
        do_chop = true;    

    bool do_jacobian = outputMode > 0;
    bool do_normal = false;

	int res = (int)pow(2.0, resolution);
    //delete (ocean->_ocean_context); 
    //delete (ocean->_ocean);
    ocean->_ocean = new drw::Ocean(res, res, size/float(res), size/float(res),
                                windSpeed, shortestWave, 0.00001f, windDirection/180.0f * M_PI,
                                1.0f - dampReflections, windAlign, oceanDepth, seed);
    ocean->_normalize_factor = ocean->_ocean->get_height_normalize_factor();
    ocean->_ocean_context = ocean->_ocean->new_context(true, do_chop, do_normal, do_jacobian);
    // sum up the waves at this timestep
    ocean->_ocean->update(time, *ocean->_ocean_context, true, do_chop, do_normal, do_jacobian, ocean->_normalize_factor * waveHeight, choppiness);

	VR::ProgressCallback * progress = vray->getSequenceData().progress;
	if (progress)
	{
		progress->info(" ");
		progress->info("HotOcean ");
		progress->info("-------------------------");
		progress->info("DeformSpace = %i", deformSpace);
		progress->info("OutputMode = %i", outputMode);
		progress->info("Normalisation factor = %f", ocean->_normalize_factor);
		progress->info("Time = %f", time);
	}
}

void TexHotOcean::frameEnd(VR::VRayRenderer *vray) {
	VRayTexture::frameEnd(vray);
	shadeCache.frameEnd(vray);
	delete (ocean->_ocean_context); 
    delete (ocean->_ocean);
}

void TexHotOcean::writeElements(const VR::VRayContext &rc, const TexmapCache &res) {
	if (!rc.mtlresult.fragment) return;

	rc.mtlresult.fragment->setChannelDataByAlias(REG_CHAN_VFB_REFLECT, &res.Eminus);
	rc.mtlresult.fragment->setChannelDataByAlias(REG_CHAN_VFB_REFLECTION_FILTER, &res.Eplus);
	rc.mtlresult.fragment->setChannelDataByAlias(REG_CHAN_VFB_RAW_REFLECTION, &res.jMinus);
	rc.mtlresult.fragment->setChannelDataByAlias(REG_CHAN_VFB_RAW_REFLECTION, &res.jPlus);
}

AColor TexHotOcean::getTexColor(const VRayContext &rc) {	
	TexmapCache res;
	// If the surface point is already in the cache - just return the cached value.
	if (shadeCache.getCache(rc, res)) {
		writeElements(rc, res);
		return res.color;
	}

	ObjectShadeInstance *objectShadeInstance = (ObjectShadeInstance*)GET_INTERFACE(rc.rayresult.si, EXT_OBJECT_SHADE_INSTANCE);
	VR::Vector shadingPointWS = rc.rayresult.wpoint + rc.vray->getFrameData().sceneOffset;
	//in and output seems to be world space
	VR::Vector disp = VR::Vector(0, 0, 0);
	VR::Vector inDisp = VR::Vector(0, 0, 0);
	if (inDisplacement){
		VR::AColor temp = inDisplacement->getTexColor(rc);
		inDisp.x = temp.color.r;
		inDisp.y = temp.color.g;
		inDisp.z = temp.color.b;
	}

	drw::EvalData evaldata;
	switch(deformSpace)
    {
    case 0: 
        {		
		ocean->_ocean_context->eval2_xz(oneOverGlobalScale*(shadingPointWS.x + inDisp.x), oneOverGlobalScale*(shadingPointWS.z + inDisp.z), evaldata);
        disp.x = evaldata.disp[0] * globalScale;
        disp.y = evaldata.disp[1] * globalScale;
        disp.z = evaldata.disp[2] * globalScale;
        //simple approximation to get the point before deformation to calculate the utility channels
        if (outputMode > 0) 
			ocean->_ocean_context->eval2_xz(oneOverGlobalScale*(shadingPointWS.x - disp.x), oneOverGlobalScale*(shadingPointWS.z- disp.z), evaldata);
        break;
        }
    case 1: 
        {
		VR::Vector shadingPointOs = rc.rayresult.wpoint;
		VR::Vector inDispOs = inDisp;
		if (objectShadeInstance)
		{
			shadingPointOs = objectShadeInstance->worldToObjectPt(rc, rc.rayresult.wpoint);
			inDispOs = objectShadeInstance->worldToObjectPt(rc, inDispOs);
		}
		ocean->_ocean_context->eval2_xz(oneOverGlobalScale*(shadingPointOs.x), oneOverGlobalScale*(shadingPointOs.z), evaldata);
        disp.x = evaldata.disp[0] * globalScale;
        disp.y =  evaldata.disp[1] * globalScale;
        disp.z = evaldata.disp[2] * globalScale;		
        //simple approximation to get the point before deformation to calculate the utility channels
        if (outputMode > 0) 
			ocean->_ocean_context->eval2_xz(oneOverGlobalScale*(shadingPointOs.x - disp.x), oneOverGlobalScale*(shadingPointOs.z - disp.z), evaldata);
		disp = objectShadeInstance->objectToWorldVec(rc, disp);
        break;		
        }		
    case 2:
        {
		MappedSurface *surf = (MappedSurface*)GET_INTERFACE(rc.rayresult.sd, EXT_MAPPED_SURFACE);
		Matrix worldToTex = Matrix(1);
		VR::Vector uvw = rc.rayresult.wpoint;
		if (surf)
		{
			worldToTex = surf->getLocalUVWTransform(rc, 0).m; // this transforms from world space to texture space
			uvw = surf->getLocalUVWCoordinates(rc, 0);
		}

		Matrix texToWorld = inverse(worldToTex);
		VR::Vector tangent = texToWorld.f[0]; // this is the tangent
		VR::Vector biNormal = texToWorld.f[1]; // this is the bi-tangent 
		ocean->_ocean_context->eval2_xz(oneOverGlobalScale*uvw.x, oneOverGlobalScale*uvw.y, evaldata);
		tangent.makeNormalized0();
		biNormal.makeNormalized0();
        disp = evaldata.disp[0] * globalScale * tangent;
		disp += evaldata.disp[1] * globalScale * rc.rayresult.normal;
        disp += evaldata.disp[2] * globalScale * biNormal;

        break;
        }
    }

	
	res.jMinus.r = evaldata.Jminus;
    res.jMinus.g = evaldata.Jminus;
    res.jMinus.b = evaldata.Jminus;
	res.jPlus.r = evaldata.Jplus;
    res.jPlus.g = evaldata.Jplus;
    res.jPlus.b = evaldata.Jplus;
	res.Eminus.r = evaldata.Eminus[0];
    res.Eminus.g = evaldata.Eminus[1];
    res.Eminus.b = evaldata.Eminus[2];
    res.Eplus.r = evaldata.Eplus[0];
    res.Eplus.g = evaldata.Eplus[1];
    res.Eplus.b = evaldata.Eplus[2];

    switch(outputMode)
    {
    case 0:
        //disp += inDisp;
		res.color.color.r = disp.x + inDisp.x;
		res.color.color.g = disp.y + inDisp.y;
		res.color.color.b = disp.z + inDisp.z;
        break;
    case 1:
		res.color = res.jMinus;
        break;
    case 2:
		res.color = res.jPlus;
        break;
    case 3:
		res.color = res.Eminus;
        break;
    case 4:
		res.color = res.Eplus;
        break;
    }

	shadeCache.putCache(rc, res);
	writeElements(rc, res);
	return res.color;
}

Vector TexHotOcean::getColorBumpGradient(const VRayContext &rc) {
	return Vector(0.0f, 0.0f, 0.0f);
}

void TexHotOcean::getTexColorBounds(AColor &cmin, AColor &cmax) {
	cmin.color.set(-1e6f, -1e6f, -1e6f); 
	cmax.color.set(1e6f, 1e6f, 1e6f);
}

int TexHotOcean::getSource(const char *part, const char *outParam, const char *lang, VUtils::ShaderSource &src, OCLData &mtlData)
{
	return 0;
}

